/*****************************************************************************\
 * TP CPE, 3ETI, TP synthese d'images
 * --------------
 *
 * Programme principal des appels OpenGL
\*****************************************************************************/



#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cmath>

#define GLEW_STATIC 1
#include <GL/glew.h>
#include <GL/glut.h>

#include "glutils.hpp"
#include "vec2.hpp"
#include "vec3.hpp"
#include "triangle_index.hpp"
#include "image.hpp"


/*****************************************************************************\
 * Structure
 *
 *
\*****************************************************************************/

//structure d'un vertex OpenGL contenant une position, une normale, une couleur et une texture
struct vertex_opengl
{
    vec3 position;
    vec3 normal;
    vec3 color;
    vec2 texture;
};



/*****************************************************************************\
 * Variables globales
 *
 *
\*****************************************************************************/

//rotation de la camera
int rotation_x=0;
int rotation_y=0;

//translation de l'objet
vec3 translation;

//identifiant du shader
GLuint shader_program_id;

//identifiant texture
GLuint texture_id;

//vertex buffer
GLuint vbo,vboi;

//temps
int counter_time=0;




/*****************************************************************************\
 * Fonctions GLUT
 *
 *
\*****************************************************************************/

//Fonction d'affichage
static void display_callback()
{
    //mise en place de la matrice de projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30,1.0,0.1,20.0); //une matrice de perspective (affiche profondeur z entre [0.1,20])

    //mise en place de la matrice "monde"
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0f,0.0f,-4.0f); //deplacement de la camera vers l'arriere
    glRotatef(rotation_x,1.0f,0.0f,0.0f); //rotation suivant un axe
    glRotatef(rotation_y,0.0f,1.0f,0.0f); //rotation suivant l'autre axe


    //effacement des couleurs du fond d'ecran
    //glClearColor(0.5, 0.6, 0.9, 1.0);
    glClearColor(1, 1, 1, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);




    //Placement du vbo courant (ici il n'y a qu'un seul vbo: optionnel)
    glBindBuffer(GL_ARRAY_BUFFER,vbo); PRINT_OPENGL_ERROR();

    //Placement du pointeur pour les positions dans le buffer de donnees
    glEnableClientState(GL_VERTEX_ARRAY); PRINT_OPENGL_ERROR();
    glVertexPointer(3, GL_FLOAT, sizeof(vertex_opengl), 0); PRINT_OPENGL_ERROR();

    //Placement du pointeur pour les normales dans le buffer de donnees
    glEnableClientState(GL_NORMAL_ARRAY); PRINT_OPENGL_ERROR();
    glNormalPointer(GL_FLOAT, sizeof(vertex_opengl),buffer_offset(sizeof(vec3))); PRINT_OPENGL_ERROR();

    //Placement du pointeur pour les couleurs dans le buffer de donnees
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3,GL_FLOAT,sizeof(vertex_opengl),buffer_offset(2*sizeof(vec3)));

    //Placement du pointeur pour les coordonnees de textures dans le buffer de donnees
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glTexCoordPointer(2, GL_FLOAT, sizeof(vertex_opengl), buffer_offset(3*sizeof(vec3)));  PRINT_OPENGL_ERROR();

    //Placement du vbo d'indice courant (ici il n'y a qu'un seul vboi: optionnel)
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,vboi);


    /****** Affichage objet 1 ******/
    glUniform3f(get_uni_loc(shader_program_id, "translation"), translation.x,translation.y,translation.z); PRINT_OPENGL_ERROR();
    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_INT, 0); PRINT_OPENGL_ERROR();

    /****** Affichage objet 2 ******/
    glUniform3f(get_uni_loc(shader_program_id, "translation"), 0.5f,0.0f,0.2f); PRINT_OPENGL_ERROR();
    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_INT, 0); PRINT_OPENGL_ERROR();



    //Changement de buffer d'affichage pour eviter un effet de scintillement
    glutSwapBuffers();
    PRINT_OPENGL_ERROR(); //recherche d'erreurs opengl
}

/*****************************************************************************\
 * init                                                                      *
\*****************************************************************************/
static void init()
{

    //Un tableau de vertex_opengl (voir definition en haut de fichier)
    vertex_opengl vertex[3];



    //mise en place des positions
    vertex[0].position=vec3(0,0,0);
    vertex[1].position=vec3(1,0,0);
    vertex[2].position=vec3(0,1,0);

    //mise en place des normales
    vertex[0].normal=vec3(0,0,1);
    vertex[1].normal=vec3(0,0,1);
    vertex[2].normal=vec3(0,0,1);

    //mise en place des couleurs
    vertex[0].color=vec3(1,0.6,0.6);
    vertex[1].color=vec3(0.6,1,0.6);
    vertex[2].color=vec3(0.6,0.6,1);

    //mise en place des textures
    vertex[0].texture=vec2(0,0);
    vertex[1].texture=vec2(1,0);
    vertex[2].texture=vec2(0,1);


    //Les indices des sommets du maillage (ici un seul triangle)
    triangle_index index[1];
    index[0]=triangle_index(0,1,2);


    //Envoie des donnees de "vertex" en memoire video
    glGenBuffers(1,&vbo); PRINT_OPENGL_ERROR();
    glBindBuffer(GL_ARRAY_BUFFER,vbo); PRINT_OPENGL_ERROR();
    glBufferData(GL_ARRAY_BUFFER,sizeof(vertex),vertex,GL_STATIC_DRAW); PRINT_OPENGL_ERROR();

    //Envoie des indices en memoire video
    glGenBuffers(1,&vboi); PRINT_OPENGL_ERROR();
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,vboi);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,sizeof(index),index,GL_STATIC_DRAW);


    // Chargement du shader
    shader_program_id = read_shader("shader.vert", "shader.frag");PRINT_OPENGL_ERROR();


    // Chargement d'une texture (seul les textures tga sont supportes)
    Image  *image = image_load_tga("texture.tga");
    if (image) //verification que l'image est bien chargee
    {

        //Creation d'un identifiant pour la texture
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glGenTextures(1, &texture_id);

        //Selection de la texture courante a partir de son identifiant
        glBindTexture(GL_TEXTURE_2D, texture_id);

        //Parametres de la texture
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

        //Envoie de l'image en memoire video
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->width, image->height, 0, GL_RGB, GL_UNSIGNED_BYTE, image->data);

        delete image;
    }
    else
    {
        std::cerr<<"Erreur chargement de l'image, etes-vous dans le bon repertoire?"<<std::endl;
        abort();
    }
    glUniform1i (get_uni_loc(shader_program_id, "texture"), 0); PRINT_OPENGL_ERROR();


    //activation de la gestion de la profondeur
    glEnable(GL_DEPTH_TEST); PRINT_OPENGL_ERROR();

}

/*****************************************************************************\
 * keyboard_callback                                                         *
\*****************************************************************************/
static void keyboard_callback(unsigned char key, int, int)
{
    //quitte le programme si on appuis sur les touches 'q', 'Q', ou 'echap'
    switch (key)
    {
    case 'Q':
    case 27:
        exit(0);

    case 'z':
        translation.z-=0.1f;
        break;
    case 's':
        translation.z+=0.1f;
        break;
    case 'q':
        translation.x-=0.1f;
        break;
    case 'd':
        translation.x+=0.1f;
        break;
    case 'r':
        translation.y+=0.1f;
        break;
    case 'f':
        translation.y-=0.1f;
        break;

    }
}

/*****************************************************************************\
 * special_callback                                                          *
\*****************************************************************************/
static void special_callback(int key, int,int)
{
  switch (key)
  {
  case GLUT_KEY_UP:
      rotation_x+=3; //rotation avec la touche du haut
      break;
  case GLUT_KEY_DOWN:
      rotation_x-=3; //rotation avec la touche du bas
      break;
  case GLUT_KEY_LEFT:
      rotation_y+=3; //rotation avec la touche de gauche
      break;
  case GLUT_KEY_RIGHT:
      rotation_y-=3; //rotation avec la touche de droite
      break;
  }

  //reactualisation de l'affichage
  glutPostRedisplay();
}

/*****************************************************************************\
 * timer_callback                                                            *
\*****************************************************************************/
static void timer_callback(int)
{
    //demande de rappel de cette fonction dans 25ms
    glutTimerFunc(25, timer_callback, 0);

    counter_time++;

    //reactualisation de l'affichage
    glutPostRedisplay();
}

int main(int argc, char** argv)
{
    //**********************************************//
    //Lancement des fonctions principales de GLUT
    //**********************************************//

    //initialisation
    glutInit(&argc, argv);

    //Mode d'affichage (couleur, gestion de profondeur, ...)
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

    //Taille de la fenetre a l'ouverture
    glutInitWindowSize(800, 800);

    //Titre de la fenetre
    glutCreateWindow("OpenGL");

    //Fonction de la boucle d'affichage
    glutDisplayFunc(display_callback);

    //Fonction de gestion du clavier
    glutKeyboardFunc(keyboard_callback);

    //Fonction des touches speciales du clavier (fleches directionnelles)
    glutSpecialFunc(special_callback);

    //Fonction d'appel d'affichage en chaine
    glutTimerFunc(25, timer_callback, 0);

    //Initialisation des fonctions OpenGL
    glewInit();

    //Notre fonction d'initialisation des donnees et chargement des shaders
    init();


    //Lancement de la boucle (infinie) d'affichage de la fenetre
    glutMainLoop();

    //Plus rien n'est execute apres cela

    return 0;
}

